#include <ESP8266WiFi.h>
#include <ESP8266HTTPClient.h>
#include <WiFiClient.h>
#include <ESP8266WebServer.h>
#include <ESP8266mDNS.h>

#include <WiFiUdp.h>
#include <ArduinoOTA.h>
#include <Time.h>
#include <SoftwareSerial.h>

/*        
                                ______________________
                                |       |___|        |
                  VCC RS485 3v3 |                    | 5v  
                            15  | D8   ESP           | GND  
    (auf TX!!) SSerial  TX  13  | D7   8266       D4 | 2 
							12  | D6              D3 | 0 
    (auf RX!!) SSerial  RX  14  | D5              D2 | 4 
							16  | D0              D1 | 5 
                         Analog | A0              RX | 3 USB
                          Reset | RST             TX | 1 USB
                                |____________________|

*/


	
	// ARRAYS
	#include "my_arrays.h"
	#include "var_reg.h"
	#include "make_fnc.h"
	#include "LittleFS.h"

// ***************************************************************
// ********************* SETUP    *************************
// ***************************************************************	
	void setup() {

		Serial.begin(115200);
		swSer.begin(9600);
		Serial.println("\nStart Uart .. ");
		
		// Wifi start
		if ( WiFi.status() != WL_CONNECTED ) starte_wifi();
		
		// Time
		configTime(MY_TZ, NTP_SERVER[0]); // --> Here is the IMPORTANT ONE LINER needed in your sketch!
		//sntp_startup_delay_MS_rfc_not_less_than_60000 ();
		
		time(&now);                       // read the current time
		localtime_r(&now, &tm);           // update the structure tm with the current time

		// Server starten
		//server_intern.on ( "/", handleRoot );
		server_intern.on ( "/json", handleJson );
		server_intern.on ( "/json_energy", handleJson_energy );
		server_intern.on ( "/ajax", handleAjax );
		server_intern.on ( "/set", handleSET );
		server_intern.on ( "/json_bat", handleBat );
		server_intern.on ( "/log", handleLog );
		server_intern.on ( "/test", handleTest );
		server_intern.enableCORS(1);; //This is the magic
		server_intern.begin();
		Serial.println ("Server gestartet"); 
		
		setupFS();
		
		// OTA
		ArduinoOTA.setHostname(DNS);
		ArduinoOTA.begin();
		Serial.println ("OTA gestartet"); 
		
	}
	

// ***************************************************************
// ********************* LOOP   *************************
// ***************************************************************	
	void loop() {
		
		if ( WiFi.status() != WL_CONNECTED ) starte_wifi(); // Wifi checken
		server_intern.handleClient();    // *** webserver
		delay(20);
		ArduinoOTA.handle();      // *** OTA
		delay(20);
		// time ***********
		localtime_r(&now, &tm);           // update the structure tm with the current time
		kucke_auf_uhr();
		
		if (swSer.available()) get_data(); // get data ***

		if ( time(&now) > (T1_alt + intervall_send) ) { // sende cmd
			swSer.write(cmd_get[loop_i],10);
			loop_i++;
			T1_alt = time(&now);
		}
		if (loop_i > 5) loop_i = 0;	// alle Reg durch
		delay(10);
		if (swSer.available()) get_data(); // get data ***
		
		if ( time(&now) > (T2_alt + intervall_json) ) { // make json
			get_http("sol_rad");
			make_json_side();
			make_json_bat();
			make_json_energy();
			make_json_DB();
			T2_alt = time(&now);
		}
		
		if ( time(&now) > (T3_alt + intervall_ajax) ) {// make ajax
			Serial.println("make json");
			get_values();
			make_json_side();	// ajax
			T3_alt = time(&now);
		}
		

		// **** Neue Anfangswerte Mitternacht ***********
		if (tm.tm_mday != alt_tag ) {
			check_suntime = false;
			sunrise = 0; sunset = 0;
			Load_max = 0.00;
			PV_max = 0.00;
			if ((tm.tm_mday*1) == 1) config_timestamp_neu(); // am Monatsanfang
			alt_mon = tm.tm_mon;
			alt_tag = tm.tm_mday;
			check_anfang_werte = false;
		}

		if (!check_suntime || !sunrise) {
			get_http("sunrise");
			get_http("sunset");
			if (sunrise > 0 && sunset > 0 ) check_suntime = true;
		}
		
		if (log_txt.length() > 1000) log_txt = "";
		if (test_out.length() > 1000) test_out = "";

	} // loop

// ***************************************************************
// ********************* SEND CMD *******************************
// ***************************************************************
	void send_cmd() {	// nach index
		swSer.write(cmd[cmd_num],lenarr[cmd_num]);
		delay(10);
	}
// ***************************************************************
// ********************* SEND FREE CMD ***************************
// ***************************************************************
	void send_free_cmd(String mystring) {
		// chars
		const char *delimiter = ",";
		char *token;
		uint16_t len = mystring.length()+1;
		char buf[len];

		// num
		byte val[10], len_arr=0;

		// String to chararr
		mystring.toCharArray(buf, len);
		Serial.print("  buf: ");
		Serial.println(buf);
		// splitt 
		token = strtok(buf, delimiter);
		while (token != NULL) {
			// covert to byte
			val[len_arr] = atoi(token);
			Serial.println(token);
			// len arr
			len_arr++;
			// nächster token
			token=strtok(NULL, delimiter);
		} // alle bytes durch
		Serial.print("len_byte-arr: ");
		Serial.print(len_arr);
		
		swSer.write(val,len_arr);
		delay(10);
	}
// ***************************************************************
// ********************* SEND BAT CONF CMD ***********************
// ***************************************************************
// [0xaa, 0x01, 0x00, 0x02, 0x07, 0x04, offset[i], 0x00, lsb, 0x00, check];
	void send_cmd_bat(byte off, byte lsb, byte che) {
		byte cbat[11] = {0xaa, 0x01, 0x00, 0x02, 0x07, 0x04, 0x55, 0x00, 0x55, 0x00, 0x55};
		cbat[6] = off;
		cbat[8] = lsb;
		cbat[10] = che;
		swSer.write(cbat,11);
		get_check_arr = 0;
		delay(10);
	}

// ***************************************************************
// ********************* GET DATA  *******************************
// ***************************************************************
	void get_data() {	
		byte i = 0, anz = 0xff, letzte_byte = 0xff, reg = 99;
		uint32_t summe = 0;
		uint32_t zeitaus = millis()+3000; // 3sec
		byte buf[60];
		
		while (swSer.available() || (i < anz && zeitaus > millis()) ) {
			
			byte ch = (byte)swSer.read();
			buf[i] = ch;
			if (i < letzte_byte) summe += ch;
			
			if (i == 3 && ch != 3) return; // wenn resp ohne antwort
			if (i == 4) reg = ch;	// register
			if (i == 5) {
				anz = ch + 0x06 + 0x01;
				letzte_byte = anz - 1;
			}
			Serial.print(" pos:");
			Serial.print(i);
			
			if (i == letzte_byte) {
				if ( ((0-summe)& 0xff) != ch ) return;
			}
			i++;
			
			Serial.print(":0x");
			if ( ch < 10 ) Serial.print("0");
			Serial.print(ch, HEX);
		} // alle empfangen **************************
		
			Serial.println("");
		if (i >= anz) { // daten vollständig
			log_txt += "\nGet data Reg: ";
			log_txt += (String)reg;
			if ( reg < 11) {
				memcpy(regs[reg], buf, sizeof(buf)); // status
				got_reg[reg] = 1;
			}
			if ( reg == 0x04) { // LOAD
				if (buf[11] == 0x13 && buf[10] == 0xfe) load_onoff = 0;
				if (buf[11] == 0x13 && buf[10] == 0xfd) load_onoff = 1;
			}
			make_out(buf, i); // schreibe log
			make_out(regs[reg], i); // schreibe log
		}
		i = 0;
	} // fnc ende get data
	
// ***************************************************************
// ********************* MAKE OUT LOG DATAARRAYS *****************
// ***************************************************************
	void make_out(byte *data, uint8_t anz) {	
	
		// Uhr ins format
		String tmmin = ":";
		String tmsec = ":";
		if (tm.tm_min < 9) tmmin = ":0";
		if (tm.tm_sec < 9) tmsec = ":0";
		
		// head
		String json_konstrukt = "Zeit: ";
		json_konstrukt += zeit_akt;
		// datum
		//json_konstrukt += "\ndatum: ";
		//json_konstrukt += datum_akt;
		
		uint32_t summe = 0;
		byte letzte_byte = anz - 1;

		json_konstrukt += "\n";
		for (uint8_t i=0; i<anz; i++) {
			
			// check byte
			if (i < letzte_byte) summe += data[i];

			json_konstrukt += (String)data[i];
			json_konstrukt += ",";
			
			if (i == letzte_byte) {
				json_konstrukt += "\nSumme: ";
				json_konstrukt += (String)summe;
				json_konstrukt += "  chk Byte: ";
				json_konstrukt += (String)data[i];
				byte calc = (0-summe)& 0xff;
				if (calc == data[i]) {
					json_konstrukt += "  chk Okay :-) ";
				}else{
					json_konstrukt += "  chk nix gut ";
				}
			}
		}
		json_konstrukt += "\n";
		/*
		for (uint8_t i=0; i<anz; i++) {
			if (data[i] < 0x10) { json_konstrukt += "0x0"; }
			else  { json_konstrukt += "0x"; }
			json_konstrukt += String(data[i],HEX);
			json_konstrukt += ",";
		}
		*/
		respose_bytes = json_konstrukt + "\n";
		test_out += json_konstrukt;
		Serial.println(respose_bytes);

	} // ende get raw

	
		

// **********************************************************************
// +++++++++++++   SERVER    ++++++++++++++++++++++++++++++++++++++++++++
// **********************************************************************

	// Json
	void handleJson() {  server_intern.send ( 200, "text/plain", json_stat ); }
	void handleJson_energy() {  server_intern.send ( 200, "text/plain", json_energy ); }
	void handleAjax() {  server_intern.send ( 200, "text/plain", ajax ); }
  
	// root
	// void handleRoot() {  server_intern.send ( 200, "text/html", index_html ); }
	
	// Battery param
	void handleBat() {  server_intern.send ( 200, "text/plain", json_bat ); }
	
	// data raw 30 sec
	void handleTest() {  server_intern.send ( 200, "text/plain", test_out ); }
	// data raw 30 sec
	void handleLog() {  server_intern.send ( 200, "text/plain", log_txt ); }
	
	
// **********************************************************************
// +++++++++++++   Client  ++++++++++++++++++++++++++++++++++++++++++++
// **********************************************************************
	void get_http( String key ) {
		
		String payload = "";
		String 	anhang = "need=" + key;
		String url = url_get + anhang;	
		
		WiFiClient client;
		HTTPClient http;
		Serial.println("[HTTP] begin...");
		Serial.println(url);
    
		if (http.begin(client, url)) { //HTTP

			int httpCode = http.GET();
			if(httpCode > 0) {
				if(httpCode == HTTP_CODE_OK) {
					payload = http.getString();
					Serial.print("payload: ");
					Serial.println(payload);

					if (  payload.length() > 0){  // wenn korrekt
						
						if (key == "sunrise") sunrise = payload.toInt();
						if (key == "sunset") sunset = payload.toInt();
						if (key == "sol_rad") sol_rad = payload.toInt();

						
					} // wenn was da
				} // if httpCode
			}
			http.end();
		} else {
		  Serial.println("[HTTP} URL nicht erreichbar");
		}
	} // ende fnc
	
	// ****************************************************************
	// ************** HANDLE SET **************************************
	// ****************************************************************
	void handleSET() { 
	
		uint8_t goffset=0, glsb=0, gcheck=0; // für Bat cmd
		get_check_arr = 0;
		
		String txt_back = "";

		//message += server_intern.uri(); ** Aufruf Verzeichnis
		// message += ( server_intern.method() == HTTP_GET ) ? "GET" : "POST"; ** CHECK GET - POST
		Serial.print( "Anzahl Args (GET): ");
		Serial.println( server_intern.args()); //	** ANZAHL GET

		for ( uint8_t i = 0; i < server_intern.args(); i++ ) { // Alle durch
			String tmp = String (i) + " " + server_intern.argName(i) + ": " + server_intern.arg(i) + "\n";
			Serial.println(tmp);

			// reset
			if (server_intern.argName(i) == "reset" && server_intern.arg(i) == "1") {
				server_intern.send ( 200, "text/plain", "reset ESP - please wait !" );
				delay(500);
				ESP.reset();
			}else{
				
				
				// Bat conf protec ***********
				if 	(server_intern.argName(i) == "bat_byte_offset") { // 3 Bytes
					goffset = server_intern.arg(i).toInt(); 
					get_check_arr++;
					txt_back += server_intern.argName(i) + "-" + server_intern.arg(i) + "\n";
				}else if (server_intern.argName(i) == "bat_byte_lsb") { // 3 Bytes
					glsb = server_intern.arg(i).toInt(); 
					get_check_arr++;
					txt_back += server_intern.argName(i) + "-" + server_intern.arg(i) + "\n";
				}else if (server_intern.argName(i) == "bat_byte_check") { // 3 Bytes
					gcheck = server_intern.arg(i).toInt(); 
					get_check_arr++;
					txt_back += server_intern.argName(i) + "-" + server_intern.arg(i) + "\n";

				// Serveraddresse
				}else if (server_intern.argName(i) == "url_get") { // default "http://192.168.123.1/datafiles/set_pv_onoff.php"
					url_get = server_intern.arg(i);
					txt_back += server_intern.argName(i) + "-" + server_intern.arg(i) + "\n";

				// Serveraddresse html Bibliotheken
				}else if (server_intern.argName(i) == "server") { // default server = "http://192.168.123.6/libs_intern_wan/webseiten";
					server = server_intern.arg(i);
					txt_back += server_intern.argName(i) + "-" + server_intern.arg(i) + "\n";

				// sol_rad
				}else if (server_intern.argName(i) == "sol_rad") { // default 
					sol_rad = server_intern.arg(i).toInt();
					txt_back += server_intern.argName(i) + "-" + server_intern.arg(i) + "\n";

				// cmd_index
				}else if (server_intern.argName(i) == "cmd_index") { // index cmd 
					cmd_num = server_intern.arg(i).toInt();
					send_cmd();
					txt_back += server_intern.argName(i) + "-" + server_intern.arg(i) + "\n";
				
				// free cmd string dec komaseperate
				}else if (server_intern.argName(i) == "cmd_free") { // index cmd 
					String comand_str = server_intern.arg(i);
					send_free_cmd(comand_str);
					txt_back += server_intern.argName(i) + "-" + server_intern.arg(i) + "\n";

				// wifi_reset_en
				}else if (server_intern.argName(i) == "wifi_reset_en") {  
				  if (server_intern.arg(i) == "1") wifi_reset_en = true; 
				  if (server_intern.arg(i) == "0") wifi_reset_en = false;
				  txt_back += server_intern.argName(i) + "-" + server_intern.arg(i) + "\n";
				  
				// load ON/OFF
				}else if (server_intern.argName(i) == "load_onoff") {  
				  if (server_intern.arg(i) == "1") cmd_num = 1; 
				  if (server_intern.arg(i) == "0") cmd_num = 0;
				  send_cmd();
				  txt_back += server_intern.argName(i) + "-" + server_intern.arg(i) + "\n";
          
				// sunrise
				}else if (server_intern.argName(i) == "sunrise") { // stamp
					sunrise = server_intern.arg(i).toInt();
					txt_back += server_intern.argName(i) + "-" + server_intern.arg(i) + "\n";
				
				// sunset
				}else if (server_intern.argName(i) == "sunset") { // stamp
					sunset = server_intern.arg(i).toInt();
					txt_back += server_intern.argName(i) + "-" + server_intern.arg(i) + "\n";
				}
				
			} // nicht reset
		} // ** alle args durch **
		
		// sende Bat conf
		if (get_check_arr == 3) send_cmd_bat(goffset, glsb, gcheck);
		
		server_intern.send ( 200, "text/plain", txt_back );

	} // ende fcn

// ***************************************************************
// ********************* KUCK AUF UHR *************************
// ***************************************************************
	void kucke_auf_uhr() {
		
		// Uhr ins format
		String tmmin = ":";
		String tmsec = ":";
		if (tm.tm_min < 9) tmmin = ":0";
		if (tm.tm_sec < 9) tmsec = ":0";

		// zeit
		zeit_akt = (String)tm.tm_hour + tmmin +(String)tm.tm_min+ tmsec +(String)tm.tm_sec;
		// datum
		datum_akt = (String)tm.tm_mday + "." +(String)(tm.tm_mon+1)+"."+(String)(tm.tm_year+1900);
		
		// Prüfe ob TAG oder NACHT
		if (sunset > 0 && sunrise > 0) {
			if (time(&now) > sunrise && time(&now) < sunset) {
				daylight = true;
			}else{
				daylight = false;
			}
		}
		
	}


// ***************************************************************
// ********************* CHECK BYTE   *************************
// ***************************************************************	
	byte Checkbyte (byte *pbData, uint16_t uwLen)	{
		byte ubChksum = 0;
		while(uwLen--)	{
			ubChksum += *pbData++;
		}
		return (0 - ubChksum) & 0xFF;
	} // checkbyte

// ***************************************************************
// ********************* SET TIMESTAMP *************************
// ***************************************************************
	void config_timestamp_neu() {
		
		// Uhr ins format
		String tmmin = ":";
		String tmsec = ":";
		if (tm.tm_min < 9) tmmin = ":0";
		if (tm.tm_sec < 9) tmsec = ":0";
		
		// stamp
		Serial.print( "\"PV_zeitstamp\":");
		Serial.println(time(&now));
		// zeit
		Serial.print("PV_zeit: ");
		uint16_t h = tm.tm_hour;
		uint16_t min = tm.tm_min;
		uint16_t sec = tm.tm_sec;
		Serial.println( (String)h + tmmin +(String)min+ tmsec +(String)sec);
		// datum
		Serial.print("PV_datum: ");
		uint16_t tag = tm.tm_mday;
		uint16_t mon = tm.tm_mon+1;
		uint16_t jahr = tm.tm_year+1900;
		Serial.print( (String)tag +"."+(String)mon +"."+(String)jahr );
		/*
			Time correction：
			Write data：AA 01 00 02 06 10 01 00 04 00 01 00 01 00 01 00 02 00 02 00 01 00 30
			Description：
			AA Data packet header
			01 Device Type: MPPT Controller
			00 Address: 0 broadcast address
			02 Command：CMD_SET
			06 Data item：remoteControl
			10 Data region length： (dec 16)
			01 00 Offset address - Remote command
			04 00 Subcommand
			01 00 Year
			01 00 Month
			01 00 Day
			02 00 Hour
			02 00 Minute
			01 00 Second
			30 checksum
			
				RemoteControl information
				Data item ID: 0x06
				Attributes： Read and write
				Bytes   Content    Offset Unit              Remarks
				2 int16 uwMagicNum; 00 00 ---------------
				2 int16 eRemoteCmd; 01 00 Remote command	(subcommand)
					typedef enum {
						0	remoteCmd_null, //  0  Meaningless
						1	remoteCmd_start, //Start charging command
						2	remoteCmd_shutdown, //Stop charging command
						3	remoteCmd_update, //Update command
						4	remoteCmd_modifyDateTime, //Synchro time command
						5	remoteCmd_startSelfTest,
						6	remoteCmd_resetAll, //Restore factory settings
						7	remoteCmd_Set
					}remoteCmdType_t;
				16 int16 uwData[8]; 02 00 Command parameters (used for time correction)
				
				strCalendarTmp.IccYear = stRemoteControl.uwData[0]; 2xByte
				strCalendarTmp.IccMon = stRemoteControl.uwData[1]; 2xByte
				strCalendarTmp.IccDay = stRemoteControl.uwData[2]; 2xByte
				strCalendarTmp.IccHours = stRemoteControl.uwData[3]; 2xByte
				strCalendarTmp.IccMin = stRemoteControl.uwData[4]; 2xByte
				strCalendarTmp.IccSec = stRemoteControl.uwData[5]; 2xByte
				
				Respose data bytes: 170,1,0,0,6,0,79, END

				02,06,10, 01,00,04,00,e7,07,08,00,06,00,0c,00,00,00,00,00,
		*/

		// make cmd SET
		//{ 0xaa, 0x01, 0x00, 0x02, 0x06, 0x10, 0x01, 0x00, 0x04, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x02, 0x00, 0x02, 0x00, 0x01, 0x00, 0x30 }
		byte cmd[23] = { 0xaa, 0x01, 0x00, 0x02, 0x06, 0x10, 0x01, 0x00, 0x04, 0x00,  // 0-9
		0xb1, 0x07,  // 10 Jahr
		0x08, 0x00,  // 12 Mon
		0x06, 0x00,  // 14 Tag 
		0x0c, 0x00,  // 16 H
		0x00, 0x00,  // 18 min
		0x00, 0x00,  // 20 sec
		0x66 };		 // 22 check
		
		cmd[10] = jahr & 0xFF;  	cmd[11] = (jahr >> 8) & 0xFF; 	// ** Jahr 10 - LSB 11 - MSB
		cmd[12] = mon & 0xFF;  		cmd[13] = (mon >> 8) & 0xFF; 	// ** Monat
		cmd[14] = tag & 0xFF;  		cmd[15] = (tag >> 8) & 0xFF; 	// ** Tag
		cmd[16] = h & 0xFF;  		cmd[17] = (h >> 8) & 0xFF; 		// ** Stunde
		cmd[18] = min & 0xFF;  		cmd[19] = (min >> 8) & 0xFF; 	// ** Minute
		cmd[20] = sec & 0xFF;  		cmd[21] = (sec >> 8) & 0xFF; 	// ** Sekunde
		
		// check extrik
		uint32_t sum = 0;
		for (byte i=0; i<22; i++) { sum += cmd[i]; }
		cmd[22] = (0 - sum) & 0xff;
		
		swSer.write(cmd,23); // sende time
		delay(10);
		
	} // ende SET


// **********************************************************************
// +++++++++++++   Wifi ++++++++++++++++++++++++++++++++++++++++++++
// **********************************************************************

	bool starte_wifi() {  
    
		//if ( WiFi.status() == WL_CONNECTED ) WiFi.mode(WIFI_OFF);
		if (wifi_reset_en) {
		  WiFi.mode(WIFI_OFF); // Neu starten
		  delay(5000);
		}
		WiFi.persistent(false);	// verhindert ständiges schreiben der Einstellungen
		WiFi.mode(WIFI_STA); // STA Client AP access point
		//WiFi.config(IPAddress(ip_add), IPAddress(gate_add), IPAddress(sub_add));
		WiFi.hostname(DNS);
		WiFi.begin ( ssid, password );
		
		// Wait for connection
		int i_zeit = 0;
		while ( WiFi.status() != WL_CONNECTED ) {
			delay ( 500 );
			i_zeit++;
			if (i_zeit > 20) ESP.reset();
		}

		// DNS Name
		MDNS.begin(DNS);

		Serial.print ( "IP address: " );
		Serial.println (WiFi.localIP());
		Serial.print ( "RSSI: " );
		Serial.println (WiFi.RSSI() );

		return true;

	} // ende WíFi
	
